###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains a class for organizing and grouping changes to a input file
of fleur in a robust way.

Essentially a low-level version of the FleurinpModifier in aiida_fleur.

.. note::
    The docstrings for the setter methods are generated from their actual implementations
    in the :py:mod:`~masci_tools.util.xml` modules via a pre-commit hook. Changes in the docstrings
    here will be overwritten
"""
from __future__ import annotations

from typing import Any, Callable, NamedTuple
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type: ignore

from masci_tools.util.xml.collect_xml_setters import XPATH_SETTERS, SCHEMA_DICT_SETTERS, NMMPMAT_SETTERS
from masci_tools.util.xml.xml_setters_names import set_attrib_value
from masci_tools.util.xml.common_functions import clear_xml, eval_xpath_one
from masci_tools.util.schema_dict_util import ensure_relaxation_xinclude
from masci_tools.io.fleur_xml import load_inpxml
from masci_tools.util.typing import XMLFileLike, FileLike
from pathlib import Path
from lxml import etree
import warnings
#Enable warnings for missing docstrings
#pylint: enable=missing-function-docstring


class ModifierTask(NamedTuple):
    name: str
    args: tuple[Any, ...] = ()
    kwargs: dict[str, Any] = {}


class FleurXMLModifier:
    """
    Class for grouping and organizing changes to a inp.xml file of fleur via the
    xml setting methods in :py:mod:`~masci_tools.util.xml.xml_setters_names` and
    :py:mod:`~masci_tools.util.xml.xml_setters_basic`

    The basic usage is shown below

    .. code-block:: python

        from masci_tools.io.fleurxmlmodifier import FleurXMLModifier

        fmode = FleurXMLModifier()

        #Add changes by calling the methods on this class
        #(names correspond to the setting methods in the xml_setters modules)
        #They are not modifying a input file directly
        #Instead all the tasks are collected and performed in one go

        fmode.set_inpchanges({'Kmax': 4.0}) #Set Kmax to 4.0
        fmode.shift_value({'Gmax': 5.0}) #Add 5 to the current value of Gmax

        #Set the local orbital configuration on all iron atoms to '3s 3p'
        fmode.set_species('all-Fe', {'lo': [{'n':3, 'l': 's', 'type': 'SCLO'},
                                            {'n':3, 'l': 'p', 'type': 'SCLO'}]})

        #To undo the last change call undo
        #fmode.undo()

        #revert_all=True resets all added tasks
        #fmode.undo(revert_all=True)

        #To apply the changes to an input file use the modify_xmlfile method
        new_xmltree, additional_files = fmode.modify_xmlfile('/path/to/input/file/inp.xml')

    """

    _xpath_functions: dict[str, Callable] = XPATH_SETTERS
    _schema_dict_functions: dict[str, Callable] = SCHEMA_DICT_SETTERS
    _nmmpmat_functions: dict[str, Callable] = NMMPMAT_SETTERS

    _extra_functions: dict[Literal['xpath', 'schema_dict', 'nmmpmat'], dict[str, Callable]] = {}

    def __new__(cls, validate_signatures=True):

        if getattr(cls, 'xpath_functions', None) is None:
            cls.xpath_functions = {**cls._xpath_functions, **cls._extra_functions.get('xpath', {})}
            cls.schema_dict_functions = {**cls._schema_dict_functions, **cls._extra_functions.get('schema_dict', {})}
            cls.nmmpmat_functions = {**cls._nmmpmat_functions, **cls._extra_functions.get('nmmpmat', {})}

        return super().__new__(cls)

    def __init__(self, validate_signatures: bool = True) -> None:

        self._tasks: list[ModifierTask] = []
        self.validate_signatures = validate_signatures

    @classmethod
    def fromList(cls, task_list: list[tuple[str, dict[str, Any]]], *args: Any, **kwargs: Any) -> FleurXMLModifier:
        """
        Instantiate the FleurXMLModifier from a list of tasks to be added immediately

        :param task_list: list of tuples first index is the name of the method
                          second is defining the arguments by keyword in a dict

        Other arguments are passed on to the __init__ method

        :returns: class with the task list instantiated
        """

        fm = cls(*args, **kwargs)
        fm.add_task_list(task_list)
        return fm

    def add_task_list(self, task_list: list[tuple[str, dict[str, Any]]]) -> None:
        """
        Add a list of tasks to be added

        :param task_list: list of tuples first index is the name of the method
                          second is defining the arguments by keyword in a dict
        """

        facade_methods = self.get_avail_actions()

        for name, kwargs in task_list:
            try:
                facade_methods[name](**kwargs)
            except KeyError as exc:
                raise ValueError(f"Unknown modification method '{name}'") from exc

    def _validate_signature(self, name: str, *args: Any, **kwargs: Any) -> None:
        """
        DEPRECATED: use `_validate_arguments` instead without unpacking args/kwargs
        """
        warnings.warn(
            'The _validate_signature method is deprecated. '
            'Please use _validate_arguments without unpacking args/kwargs instead', DeprecationWarning)
        self._validate_arguments(name, args, kwargs)

    def _get_setter_function_and_prefix(self, name: str) -> tuple[Callable[[Any], Any], tuple[str, ...]]:
        """
        Get the setter function and a prefix standing in for the arguments that
        are substituted when performing the modification
        """
        if name in self.xpath_functions:
            func = self.xpath_functions[name]
            prefix: tuple[str, ...] = ('xmltree',)
        elif name in self.schema_dict_functions:
            func = self.schema_dict_functions[name]
            prefix = ('xmltree', 'schema_dict')
        elif name in self.nmmpmat_functions:
            func = self.nmmpmat_functions[name]
            prefix = ('xmltree', 'nmmplines', 'schema_dict')

        if func is None:
            raise ValueError(f'Failed to validate setter {name}. Maybe the function was'
                             'not registered in masci_tools.util.xml.collect_xml_setters')

        #For functions decorated with the schema_dict_version_dispatch
        #We check only the default (This function should have a compatible signature for all registered functions)
        if getattr(func, 'registry', None) is not None:
            func = func.registry['default']

        return func, prefix

    def _get_setter_func_kwargs(self, name: str, args: tuple[Any, ...], kwargs: dict[str, Any]) -> dict[str, Any]:
        """
        Map the given args and kwargs to just kwargs for the
        setter function with the given name

        :param name: name of the setter function
        :param args: positional arguments to the setter function
        :param kwargs: keyword arguments to the setter function
        """
        from inspect import signature
        func, prefix = self._get_setter_function_and_prefix(name)

        sig = signature(func)
        bound = sig.bind(*prefix, *args, **kwargs)

        kwargs_complete = {k: v for k, v in bound.arguments.items() if k not in ('xmltree', 'nmmplines', 'schema_dict')}

        #Fix if the XML modifying function has an explicit kwargs
        if 'kwargs' in kwargs_complete:
            kwargs_explicit = kwargs_complete.pop('kwargs')
            kwargs_complete = {**kwargs_complete, **kwargs_explicit}

        return kwargs_complete

    def _validate_arguments(self, name: str, args: tuple[Any, ...], kwargs: dict[str, Any]) -> None:
        """
        Validate that the given arguments to the registration
        method can be used to call the corresponding XML modifying function
        """
        from inspect import signature

        if self.validate_signatures:
            func, prefix = self._get_setter_function_and_prefix(name)
            try:
                sig = signature(func)
                sig.bind(*prefix, *args, **kwargs)
            except TypeError as exc:
                raise TypeError(
                    f"The given arguments for the registration method '{name}' are not valid for the XML modifying function"
                    f'The following error was raised: {exc}') from exc

    @classmethod
    def apply_modifications(cls,
                            xmltree: etree._ElementTree,
                            nmmp_lines: list[str] | None,
                            modification_tasks: list[ModifierTask],
                            validate_changes: bool = True,
                            adjust_version_for_dev_version: bool = True) -> tuple[etree._ElementTree, list[str] | None]:
        """
        Applies given modifications to the fleurinp lxml tree.
        It also checks if a new lxml tree is validated against schema.
        Does not rise an error if inp.xml is not validated, simple prints a message about it.

        :param xmltree: a lxml tree to be modified (IS MODIFIED INPLACE)
        :param nmmp_lines: a n_mmp_mat file to be modified (IS MODIFIED INPLACE)
        :param modification_tasks: a list of modification tuples
        :param validate_changes: bool optional (default True), if True after all tasks are performed
                                 both the xmltree and nmmp_lines are checked for consistency
        :param adjust_version_for_dev_version: bool optional (default True), if True and the schema_dict
                                               and file version differ, e.g. a development version is used
                                               the version is temporarily modified to swallow the validation
                                               error that would occur

        :returns: a modified lxml tree and a modified n_mmp_mat file
        """
        from masci_tools.util.xml.xml_setters_nmmpmat import validate_nmmpmat

        xmltree, schema_dict = load_inpxml(xmltree)
        xmltree, _ = clear_xml(xmltree)

        file_version = eval_xpath_one(xmltree, '//@fleurInputVersion', str)
        is_dev_version = schema_dict['inp_version'] != file_version

        for task in modification_tasks:
            if task.name in cls.xpath_functions:
                action = cls.xpath_functions[task.name]
                xmltree = action(xmltree, *task.args, **task.kwargs)

            elif task.name in cls.schema_dict_functions:
                action = cls.schema_dict_functions[task.name]
                xmltree = action(xmltree, schema_dict, *task.args, **task.kwargs)

            elif task.name in cls.nmmpmat_functions:
                action = cls.nmmpmat_functions[task.name]
                nmmp_lines = action(xmltree, nmmp_lines, schema_dict, *task.args, **task.kwargs)

            else:
                raise ValueError(f'Unknown task {task.name}')

        if validate_changes:
            if is_dev_version and adjust_version_for_dev_version:
                set_attrib_value(xmltree, schema_dict, 'fleurinputversion', schema_dict['inp_version'])
                schema_dict.validate(xmltree, header='Changes were not valid')
                set_attrib_value(xmltree, schema_dict, 'fleurinputversion', file_version)
            else:
                schema_dict.validate(xmltree, header='Changes were not valid')
            try:
                validate_nmmpmat(xmltree, nmmp_lines, schema_dict)
            except ValueError as exc:
                msg = f'Changes were not valid (n_mmp_mat file is not compatible): {modification_tasks}'
                raise ValueError(msg) from exc

        return xmltree, nmmp_lines

    @property
    def task_list(self) -> list[tuple[str, dict[str, Any]]]:
        """
        Return the current changes in a format accepted by :py:meth:`add_task_list()`
        and :py:meth:`fromList()`
        """

        tasks = []
        for change in self._tasks:
            #Here we already validated the arguments so we know we can just get the kwargs
            kwargs = self._get_setter_func_kwargs(change.name, change.args, change.kwargs)
            tasks.append((change.name, kwargs))

        return tasks

    def get_avail_actions(self) -> dict[str, Callable]:
        """
        Returns the allowed functions from FleurXMLModifier
        """
        outside_actions = {
            'set_inpchanges': self.set_inpchanges,
            'shift_value': self.shift_value,
            'set_xcfunctional': self.set_xcfunctional,
            'set_species': self.set_species,
            'set_species_label': self.set_species_label,
            'clone_species': self.clone_species,
            'switch_species': self.switch_species,
            'switch_species_label': self.switch_species_label,
            'shift_value_species_label': self.shift_value_species_label,
            'set_atomgroup': self.set_atomgroup,
            'set_atomgroup_label': self.set_atomgroup_label,
            'set_complex_tag': self.set_complex_tag,
            'set_simple_tag': self.set_simple_tag,
            'create_tag': self.create_tag,
            'delete_tag': self.delete_tag,
            'delete_att': self.delete_att,
            'replace_tag': self.replace_tag,
            'set_text': self.set_text,
            'set_first_text': self.set_first_text,
            'set_attrib_value': self.set_attrib_value,
            'set_first_attrib_value': self.set_first_attrib_value,
            'add_number_to_attrib': self.add_number_to_attrib,
            'add_number_to_first_attrib': self.add_number_to_first_attrib,
            'xml_create_tag': self.xml_create_tag,
            'xml_replace_tag': self.xml_replace_tag,
            'xml_delete_tag': self.xml_delete_tag,
            'xml_delete_att': self.xml_delete_att,
            'xml_set_attrib_value_no_create': self.xml_set_attrib_value_no_create,
            'xml_set_text_no_create': self.xml_set_text_no_create,
            'set_nmmpmat': self.set_nmmpmat,
            'rotate_nmmpmat': self.rotate_nmmpmat,
            'align_nmmpmat_to_sqa': self.align_nmmpmat_to_sqa,
            'set_nkpts': self.set_nkpts,
            'set_kpath': self.set_kpath,
            'set_kpointlist': self.set_kpointlist,
            'switch_kpointset': self.switch_kpointset,
            'set_kpointpath': self.set_kpointpath,
            'set_kpointmesh': self.set_kpointmesh
        }
        return outside_actions

    def undo(self, revert_all: bool = False) -> list[ModifierTask]:
        """
        Cancels the last change or all of them

        :param revert_all: set True if need to cancel all the changes, False if the last one.
        """
        if revert_all:
            self._tasks = []
        else:
            if self._tasks:
                self._tasks.pop()
                #TODO delete nodes from other nodes
                #del self._tasks[-1]
        return self._tasks

    def changes(self) -> list[ModifierTask]:
        """
        Prints out all changes currently registered on this instance
        """
        from pprint import pprint
        pprint(self._tasks)
        return self._tasks

    def modify_xmlfile(self,
                       original_inpxmlfile: XMLFileLike,
                       original_nmmp_file: FileLike | list[str] | None = None,
                       validate_changes: bool = True,
                       adjust_version_for_dev_version: bool = True) -> tuple[etree._ElementTree, dict[str, str]]:
        """
        Applies the registered modifications to a given inputfile

        :param original_inpxmlfile: either path to the inp.xml file, opened file handle
                                    or a xml etree to be parsed
        :param original_nmmp_file: path or list of str to a corresponding density matrix
                                   file

        :raises ValueError: if the parsing of the input file

        :returns: a modified xmltree and if existent a modified density matrix file
        """
        original_xmltree, schema_dict = load_inpxml(original_inpxmlfile)

        if original_nmmp_file is not None:
            if isinstance(original_nmmp_file, str) and not Path(original_nmmp_file).is_file():
                original_nmmp_lines = original_nmmp_file.split('\n')
            elif isinstance(original_nmmp_file, (str, Path)):
                with open(original_nmmp_file, encoding='utf-8') as n_mmp_file:
                    original_nmmp_lines = n_mmp_file.read().split('\n')
            else:
                original_nmmp_lines = original_nmmp_file  #type:ignore[assignment]
        else:
            original_nmmp_lines = None

        new_xmltree, new_nmmp_lines = self.apply_modifications(
            original_xmltree,
            original_nmmp_lines,
            self._tasks,
            validate_changes=validate_changes,
            adjust_version_for_dev_version=adjust_version_for_dev_version)

        ensure_relaxation_xinclude(new_xmltree, schema_dict)
        etree.indent(new_xmltree)

        additional_files = {}
        if new_nmmp_lines is not None:
            additional_files['n_mmp_mat'] = '\n'.join(new_nmmp_lines)
        return new_xmltree, additional_files

    def set_inpchanges(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_inpchanges()` to
        the list of tasks that will be done on the xmltree.

        This method sets all the attribute and texts provided in the change_dict.

        The first occurrence of the attribute/tag is set

        :param changes: dictionary {attrib_name : value} with all the wanted changes.
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

        An example of changes::

            changes = {
                'itmax' : 1,
                'l_noco': True,
                'ctail': False,
                'l_ss': True
            }
        """
        if 'change_dict' in kwargs:
            warnings.warn('The argument change_dict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('change_dict')
        self._validate_arguments('set_inpchanges', args, kwargs)
        self._tasks.append(ModifierTask('set_inpchanges', args, kwargs))

    def shift_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value()` to
        the list of tasks that will be done on the xmltree.

        Shifts numerical values of attributes directly in the inp.xml file.

        The first occurrence of the attribute is shifted

        :param changes: a python dictionary with the keys to shift and the shift values.
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                     `rel`/`relative` multiplies the old value with the given value
                     `abs`/`absolute` adds the old value and the given value
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute


        An example of changes::

                changes = {'itmax' : 1, 'dVac': -0.123}
        """
        if 'change_dict' in kwargs:
            warnings.warn('The argument change_dict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('change_dict')
        self._validate_arguments('shift_value', args, kwargs)
        self._tasks.append(ModifierTask('shift_value', args, kwargs))

    def set_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species()` to
        the list of tasks that will be done on the xmltree.

        Method to set parameters of a species tag of the fleur inp.xml file.

        :param species_name: string, name of the specie you want to change
                             Can be name of the species, 'all' or 'all-<string>' (sets species with the string in the species name)
        :param changes: a python dict specifying what you want to change.
        :param create: bool, if species does not exist create it and all subtags?
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :raises ValueError: if species name is non existent in inp.xml and should not be created.
                            also if other given tags are garbage. (errors from eval_xpath() methods)

        :return xmltree: xml etree of the new inp.xml

        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a MT radius it
        can be done via::

            changes = {'mtSphere' : {'radius' : 2.2}}

        Another example::

            'changes': {'special': {'socscale': 0.0}}

        that switches SOC terms on a sertain specie. ``mtSphere``, ``atomicCutoffs``,
        ``energyParameters``, ``lo``, ``electronConfig``, ``nocoParams``, ``ldaU`` and
        ``special`` keys are supported. To find possible
        keys of the inner dictionary please refer to the FLEUR documentation flapw.de
        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        self._validate_arguments('set_species', args, kwargs)
        self._tasks.append(ModifierTask('set_species', args, kwargs))

    def set_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species_label()` to
        the list of tasks that will be done on the xmltree.

        This method calls :func:`~masci_tools.util.xml.xml_setters_names.set_species()`
        method for a certain atom species that corresponds to an atom with a given label

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param changes: a python dict specifying what you want to change.
        :param create: bool, if species does not exist create it and all subtags?
        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        self._validate_arguments('set_species_label', args, kwargs)
        self._tasks.append(ModifierTask('set_species_label', args, kwargs))

    def clone_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.clone_species()` to
        the list of tasks that will be done on the xmltree.

        Method to create a new species from an existing one with evtl. modifications

        For reference of the changes dictionary look at :py:func:`set_species()`

        :param species_name: string, name of the specie you want to clone
                             Has to correspond to one single species (no 'all'/'all-<search_string>')
        :param new_name: new name of the cloned species
        :param changes: a optional python dict specifying what you want to change.
        """
        self._validate_arguments('clone_species', args, kwargs)
        self._tasks.append(ModifierTask('clone_species', args, kwargs))

    def switch_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_species()` to
        the list of tasks that will be done on the xmltree.

        Method to switch the species of an atom group of the fleur inp.xml file.

        :param new_species_name: name of the species to switch to
        :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
        :param species: atom groups, corresponding to the given species will be changed
        :param clone: if True and the new species name does not exist and it corresponds to changing
                      from one species the species will be cloned with :py:func:`clone_species()`
        :param changes: changes to do if the species is cloned
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        """
        self._validate_arguments('switch_species', args, kwargs)
        self._tasks.append(ModifierTask('switch_species', args, kwargs))

    def switch_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_species_label()` to
        the list of tasks that will be done on the xmltree.

        Method to switch the species of an atom group of the fleur inp.xml file based on a label
        of a contained atom

        :param atom_label: string, a label of the atom which group will be changed. 'all' to change all the groups
        :param new_species_name: name of the species to switch to
        :param clone: if True and the new species name does not exist and it corresponds to changing
                      from one species the species will be cloned with :py:func:`clone_species()`
        :param changes: changes to do if the species is cloned
        """
        self._validate_arguments('switch_species_label', args, kwargs)
        self._tasks.append(ModifierTask('switch_species_label', args, kwargs))

    def shift_value_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value_species_label()` to
        the list of tasks that will be done on the xmltree.

        Shifts the value of an attribute on a species by label
        if atom_label contains 'all' then applies to all species

        :param atom_label: string, a label of the atom which specie will be changed. 'all' if set up all species
        :param attribute_name: name of the attribute to change
        :param number_to_add: value to add or to multiply by
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                     `rel`/`relative` multiplies the old value with `number_to_add`
                     `abs`/`absolute` adds the old value and `number_to_add`

        Kwargs if the attribute_name does not correspond to a unique path:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use attribute_name instead', DeprecationWarning)
            kwargs['attribute_name'] = kwargs.pop('attributename')
        if 'value_given' in kwargs:
            warnings.warn('The argument value_given is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('value_given')
        self._validate_arguments('shift_value_species_label', args, kwargs)
        self._tasks.append(ModifierTask('shift_value_species_label', args, kwargs))

    def set_atomgroup(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup()` to
        the list of tasks that will be done on the xmltree.

        Method to set parameters of an atom group of the fleur inp.xml file.

        :param changes: a python dict specifying what you want to change.
        :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
        :param species: atom groups, corresponding to the given species will be changed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'changes': {'nocoParams': {'beta': val}}
        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        if 'create' in kwargs:
            warnings.warn('The argument create is deprecated and is ignored.', DeprecationWarning)
            kwargs.pop('create')
        self._validate_arguments('set_atomgroup', args, kwargs)
        self._tasks.append(ModifierTask('set_atomgroup', args, kwargs))

    def set_atomgroup_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup_label()` to
        the list of tasks that will be done on the xmltree.

        This method calls :func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup()`
        method for a certain atom species that corresponds to an atom with a given label.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param changes: a python dict specifying what you want to change.


        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'changes': {'nocoParams': {'beta': val}}
        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        if 'create' in kwargs:
            warnings.warn('The argument create is deprecatedand is ignored.', DeprecationWarning)
            kwargs.pop('create')
        self._validate_arguments('set_atomgroup_label', args, kwargs)
        self._tasks.append(ModifierTask('set_atomgroup_label', args, kwargs))

    def create_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.create_tag()` to
        the list of tasks that will be done on the xmltree.

        This method creates a tag with a uniquely identified xpath under the nodes of its parent.
        If there are no nodes evaluated the subtags can be created with `create_parents=True`

        The tag is always inserted in the correct place if a order is enforced by the schema

        :param tag: str of the tag to create or etree Element or string representing the XML element with the same name to insert
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param create_parents: bool optional (default False), if True and the given xpath has no results the
                               the parent tags are created recursively
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('create_tag', args, kwargs)
        self._tasks.append(ModifierTask('create_tag', args, kwargs))

    def delete_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_tag()` to
        the list of tasks that will be done on the xmltree.

        This method deletes a tag with a uniquely identified xpath.

        :param tag: str of the tag to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param occurrences: int or list of int. Which occurrence of the parent nodes to delete a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('delete_tag', args, kwargs)
        self._tasks.append(ModifierTask('delete_tag', args, kwargs))

    def delete_att(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_att()` to
        the list of tasks that will be done on the xmltree.

        This method deletes a attribute with a uniquely identified xpath.

        :param name: str of the attribute to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param occurrences: int or list of int. Which occurrence of the parent nodes to delete a attribute.
                            By default all nodes are used.

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attrib_name' in kwargs:
            warnings.warn('The argument attrib_name is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attrib_name')
        self._validate_arguments('delete_att', args, kwargs)
        self._tasks.append(ModifierTask('delete_att', args, kwargs))

    def replace_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.replace_tag()` to
        the list of tasks that will be done on the xmltree.

        This method deletes a tag with a uniquely identified xpath.

        :param tag: str of the tag to replace
        :param element: etree Element or string representing the XML element to replace the tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param occurrences: int or list of int. Which occurrence of the parent nodes to replace a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        if 'newelement' in kwargs:
            warnings.warn('The argument newelement is deprecated. Use element instead', DeprecationWarning)
            kwargs['element'] = kwargs.pop('newelement')
        self._validate_arguments('replace_tag', args, kwargs)
        self._tasks.append(ModifierTask('replace_tag', args, kwargs))

    def set_complex_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_complex_tag()` to
        the list of tasks that will be done on the xmltree.

        Function to correctly set tags/attributes for a given tag.
        Goes through the attributedict and decides based on the schema_dict, how the corresponding
        key has to be handled.
        The tag is specified via its name and evtl. further specification

        Supports:

            - attributes
            - tags with text only
            - simple tags, i.e. only attributes (can be optional single/multiple)
            - complex tags, will recursively create/modify them

        :param tag_name: name of the tag to set
        :param changes: Keys in the dictionary correspond to names of tags and the values are the modifications
                        to do on this tag (attributename, subdict with changes to the subtag, ...)
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param create: bool optional (default False), if True and the path, where the complex tag is
                       set does not exist it is created

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('set_complex_tag', args, kwargs)
        self._tasks.append(ModifierTask('set_complex_tag', args, kwargs))

    def set_simple_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_simple_tag()` to
        the list of tasks that will be done on the xmltree.

        Sets one or multiple `simple` tag(s) in an xmltree. A simple tag can only hold attributes and has no
        subtags. The tag is specified by its name and further specification
        If the tag can occur multiple times all existing tags are DELETED and new ones are written.
        If the tag only occurs once it will automatically be created if its missing.

        :param tag_name: str name of the tag to modify/set
        :param changes: list of dicts or dict with the changes. Elements in list describe multiple tags.
                        Keys in the dictionary correspond to {'attributename': attributevalue}
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param create_parents: bool optional (default False), if True and the path, where the simple tags are
                               set does not exist it is created

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('set_simple_tag', args, kwargs)
        self._tasks.append(ModifierTask('set_simple_tag', args, kwargs))

    def set_text(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_text()` to
        the list of tasks that will be done on the xmltree.

        Sets the text on tags in a xmltree to a given value, specified by the name of the tag and
        further specifications. By default the text will be set on all nodes returned for the specified xpath.
        If there are no nodes under the specified xpath a tag can be created with `create=True`.
        The text values are converted automatically according to the types
        with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
        are not `str` already.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('set_text', args, kwargs)
        self._tasks.append(ModifierTask('set_text', args, kwargs))

    def set_first_text(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_text()` to
        the list of tasks that will be done on the xmltree.

        Sets the text the first occurrence of a tag in a xmltree to a given value, specified by the name of the tag and
        further specifications. By default the text will be set on all nodes returned for the specified xpath.
        If there are no nodes under the specified xpath a tag can be created with `create=True`.
        The text values are converted automatically according to the types
        with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
        are not `str` already.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('set_first_text', args, kwargs)
        self._tasks.append(ModifierTask('set_first_text', args, kwargs))

    def set_attrib_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        Sets an attribute in a xmltree to a given value, specified by its name and further
        specifications.
        If there are no nodes under the specified xpath a tag can be created with `create=True`.
        The attribute values are converted automatically according to the types of the attribute
        with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
        are not `str` already.

        :param name: the attribute name to set
        :param value: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('set_attrib_value', args, kwargs)
        self._tasks.append(ModifierTask('set_attrib_value', args, kwargs))

    def set_first_attrib_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        Sets the first occurrence of an attribute in a xmltree to a given value, specified by its name and further
        specifications.
        If there are no nodes under the specified xpath a tag can be created with `create=True`.
        The attribute values are converted automatically according to the types of the attribute
        with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
        are not `str` already.

        :param name: the attribute name to set
        :param value: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('set_first_attrib_value', args, kwargs)
        self._tasks.append(ModifierTask('set_first_attrib_value', args, kwargs))

    def add_number_to_attrib(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_attrib()` to
        the list of tasks that will be done on the xmltree.

        Adds a given number to the attribute value in a xmltree specified by the name of the attribute
        and optional further specification
        If there are no nodes under the specified xpath an error is raised

        :param name: the attribute name to change
        :param number_to_add: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                     `rel`/`relative` multiplies the old value with `number_to_add`
                     `abs`/`absolute` adds the old value and `number_to_add`
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'add_number' in kwargs:
            warnings.warn('The argument add_number is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('add_number')
        self._validate_arguments('add_number_to_attrib', args, kwargs)
        self._tasks.append(ModifierTask('add_number_to_attrib', args, kwargs))

    def add_number_to_first_attrib(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_first_attrib()` to
        the list of tasks that will be done on the xmltree.

        Adds a given number to the first occurrence of an attribute value in a xmltree specified by the name of the attribute
        and optional further specification
        If there are no nodes under the specified xpath an error is raised

        :param name: the attribute name to change
        :param number_to_add: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                     `rel`/`relative` multiplies the old value with `number_to_add`
                     `abs`/`absolute` adds the old value and `number_to_add`
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'add_number' in kwargs:
            warnings.warn('The argument add_number is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('add_number')
        self._validate_arguments('add_number_to_first_attrib', args, kwargs)
        self._tasks.append(ModifierTask('add_number_to_first_attrib', args, kwargs))

    def xml_create_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_create_tag()` to
        the list of tasks that will be done on the xmltree.

        This method evaluates an xpath expression and creates a tag in a xmltree under the
        returned nodes.
        If there are no nodes under the specified xpath an error is raised.

        The tag is appended by default, but can be inserted at a certain index (`place_index`)
        or can be inserted according to a given order of tags

        :param xpath: a path where to place a new tag
        :param element: a tag name, etree Element or string representing the XML element to be created
        :param place_index: defines the place where to put a created tag
        :param tag_order: defines a tag order
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        :param correct_order: bool, if True (default) and a tag_order is given, that does not correspond to the given order
                              in the xmltree (only order wrong no unknown tags) it will be corrected and a warning is given
                              This is necessary for some edge cases of the xml schemas of fleur
        :param several: bool, if True multiple tags od the given name are allowed

        :raises ValueError: If the insertion failed in any way (tag_order does not match, failed to insert, ...)
        """
        self._validate_arguments('xml_create_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_create_tag', args, kwargs))

    def xml_replace_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_replace_tag()` to
        the list of tasks that will be done on the xmltree.

        Replace XML tags by a given tag on the given XML tree

        :param xpath: a path to the tag to be replaced
        :param element: an Element or string representing the Element to replace the found tags with
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        if 'newelement' in kwargs:
            warnings.warn('The argument newelement is deprecated. Use element instead', DeprecationWarning)
            kwargs['element'] = kwargs.pop('newelement')

        self._validate_arguments('xml_replace_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_replace_tag', args, kwargs))

    def xml_delete_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_tag()` to
        the list of tasks that will be done on the xmltree.

        Deletes a tag in the XML tree.

        :param xpath: a path to the tag to be deleted
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._validate_arguments('xml_delete_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_delete_tag', args, kwargs))

    def xml_delete_att(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_att()` to
        the list of tasks that will be done on the xmltree.

        Deletes an attribute in the XML tree

        :param xpath: a path to the attribute to be deleted
        :param name: the name of an attribute to delete
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        self._validate_arguments('xml_delete_att', args, kwargs)
        self._tasks.append(ModifierTask('xml_delete_att', args, kwargs))

    def xml_set_attrib_value_no_create(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_attrib_value_no_create()` to
        the list of tasks that will be done on the xmltree.

        Sets an attribute in a xmltree to a given value. By default the attribute will be set
        on all nodes returned for the specified xpath.

        :param xpath: a path where to set the attributes
        :param name: the attribute name to set
        :param value: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.

        :raises ValueError: If the lengths of attribv or occurrences do not match number of nodes
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('xml_set_attrib_value_no_create', args, kwargs)
        self._tasks.append(ModifierTask('xml_set_attrib_value_no_create', args, kwargs))

    def xml_set_text_no_create(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_text_no_create()` to
        the list of tasks that will be done on the xmltree.

        Sets the text of a tag in a xmltree to a given value.
        By default the text will be set on all nodes returned for the specified xpath.

        :param xpath: a path where to set the text
        :param text: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.

        :raises ValueError: If the lengths of text or occurrences do not match number of nodes
        """
        self._validate_arguments('xml_set_text_no_create', args, kwargs)
        self._tasks.append(ModifierTask('xml_set_text_no_create', args, kwargs))

    def set_nmmpmat(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.set_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        Routine sets the block in the n_mmp_mat file specified by species_name, orbital and spin
        to the desired density matrix

        :param species_name: string, name of the species you want to change
        :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
        :param spin: integer, specifies which spin block should be modified
        :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                                  else to zero
        :param denmat: matrix, specify the density matrix explicitly
        :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :raises ValueError: If something in the input is wrong
        :raises KeyError: If no LDA+U procedure is found on a species
        """
        self._validate_arguments('set_nmmpmat', args, kwargs)
        self._tasks.append(ModifierTask('set_nmmpmat', args, kwargs))

    def rotate_nmmpmat(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.rotate_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        Rotate the density matrix with the given angles phi and theta

        :param species_name: string, name of the species you want to change
        :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
        :param phi: float, angle (radian), by which to rotate the density matrix
        :param theta: float, angle (radian), by which to rotate the density matrix
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :raises ValueError: If something in the input is wrong
        :raises KeyError: If no LDA+U procedure is found on a species
        """
        self._validate_arguments('rotate_nmmpmat', args, kwargs)
        self._tasks.append(ModifierTask('rotate_nmmpmat', args, kwargs))

    def align_nmmpmat_to_sqa(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.align_nmmpmat_to_sqa()` to
        the list of tasks that will be done on the xmltree.

        Align the density matrix with the given SQA of the associated species

        :param species_name: string, name of the species you want to change
        :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
        :param phi_before: float or list of floats, angle (radian),
                           values for phi for the previous alignment of the density matrix
        :param theta_before: float or list of floats, angle (radian),
                             values for theta for the previous alignment of the density matrix
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :raises ValueError: If something in the input is wrong
        :raises KeyError: If no LDA+U procedure is found on a species
        """
        self._validate_arguments('align_nmmpmat_to_sqa', args, kwargs)
        self._tasks.append(ModifierTask('align_nmmpmat_to_sqa', args, kwargs))

    def set_kpointlist(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpointlist()` to
        the list of tasks that will be done on the xmltree.

        Explicitly create a kPointList from the given kpoints and weights. This routine will add the
        specified kPointList with the given name.

        .. warning::
            For input versions Max4 and older **all** keyword arguments are not valid (`name`, `kpoint_type`,
            `special_labels`, `switch` and `overwrite`)

        :param kpoints: list or array containing the **relative** coordinates of the kpoints
        :param weights: list or array containing the weights of the kpoints
        :param name: str for the name of the list, if not given a default name is generated
        :param kpoint_type: str specifying the type of the kPointList ('path', 'mesh', 'spex', 'tria', ...)
        :param special_labels: dict mapping indices to labels. The labels will be inserted for the kpoints
                               corresponding to the given index
        :param switch: bool, if True the kPointlist will be used by Fleur when starting the next calculation
        :param overwrite: bool, if True and a kPointlist with the given name already exists it will be overwritten
        """
        self._validate_arguments('set_kpointlist', args, kwargs)
        self._tasks.append(ModifierTask('set_kpointlist', args, kwargs))

    def switch_kpointset(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_kpointset()` to
        the list of tasks that will be done on the xmltree.

        Switch the used k-point set

        .. warning::
            This method is only supported for input versions after the Max5 release

        :param list_name: name of the kPoint set to use
        """
        self._validate_arguments('switch_kpointset', args, kwargs)
        self._tasks.append(ModifierTask('switch_kpointset', args, kwargs))

    def set_nkpts(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_nkpts()` to
        the list of tasks that will be done on the xmltree.

        Sets a k-point mesh directly into inp.xml

        .. warning::
            This method is only supported for input versions before the Max5 release

        :param count: number of k-points
        :param gamma: bool that controls if the gamma-point should be included
                      in the k-point mesh
        """
        self._validate_arguments('set_nkpts', args, kwargs)
        self._tasks.append(ModifierTask('set_nkpts', args, kwargs))

    def set_kpath(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpath()` to
        the list of tasks that will be done on the xmltree.

        Sets a k-path directly into inp.xml  as a alternative kpoint set with purpose 'bands'

        .. warning::
            This method is only supported for input versions before the Max5 release

        :param kpath: a dictionary with kpoint name as key and k point coordinate as value
        :param count: number of k-points
        :param gamma: bool that controls if the gamma-point should be included
                      in the k-point mesh
        """
        self._validate_arguments('set_kpath', args, kwargs)
        self._tasks.append(ModifierTask('set_kpath', args, kwargs))

    def set_kpointpath(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpointpath()` to
        the list of tasks that will be done on the xmltree.

        Create a kpoint list for a bandstructure calculation (using ASE kpath generation)

        The path can be defined explictly (see :py:func:`~ase.dft.kpoints.bandpath`) or derived from the unit cell

        :param path: str, list of str or None defines the path to interpolate (for syntax :py:func:`~ase.dft.kpoints.bandpath`)
        :param nkpts: int number of kpoints in the path
        :param density: float number of kpoints per Angstroem
        :param name: Name of the created kpoint list. If not given a name is generated
        :param switch: bool if True the kpoint list is direclty set as the used set
        :param overwrite: if True and a kpoint list of the given name already exists it will be overwritten
        :param special_points: dict mapping names to coordinates for special points to use
        """
        self._validate_arguments('set_kpointpath', args, kwargs)
        self._tasks.append(ModifierTask('set_kpointpath', args, kwargs))

    def set_kpointmesh(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpointmesh()` to
        the list of tasks that will be done on the xmltree.

        Create a kpoint mesh using spglib

        for details see :py:func:`~spglib.get_stabilized_reciprocal_mesh`

        :param mesh: list-like woth three elements, giving the size of the kpoint set in each direction
        :param use_symmetry: bool if True the available symmetry operations in the inp.xml will be used
                             to reduce the kpoint set otherwise only the identity matrix is used
        :param name: Name of the created kpoint list. If not given a name is generated
        :param switch: bool if True the kpoint list is direclty set as the used set
        :param overwrite: if True and a kpoint list of the given name already exists it will be overwritten
        :param shift: shift the center of the kpint set
        :param time_reversal: bool if True time reversal symmetry will be used to reduce the kpoint set
        :param map_to_first_bz: bool if True the kpoints are mapped into the [0,1] interval
        """
        self._validate_arguments('set_kpointmesh', args, kwargs)
        self._tasks.append(ModifierTask('set_kpointmesh', args, kwargs))

    def set_xcfunctional(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_xcfunctional()` to
        the list of tasks that will be done on the xmltree.

        Set the Exchange Correlation potential tag

        Setting a inbuilt XC functional
        .. code-block:: python

            set_xcfunctional(xmltree, schema_dict, 'vwn')

        Setting a LibXC XC functional
        .. code-block:: python

            set_xcfunctional(xmltree, schema_dict, {'exchange': 'lda_x', 'correlation':"lda_c_xalpha"}, libxc=True)

        :param xc_functional: str or dict. If str it is the name of a inbuilt XC functional. If it is a dict it
                              specifies either the name or id for LibXC functionals for the keys
                              `'exchange', 'correlation', 'etot_exchange' and 'etot_correlation'`
        :param xc_functional_options: dict with further general changes to the `xcFunctional` tag
        :param libxc: bool if True the functional is a LibXC functional
        """
        self._validate_arguments('set_xcfunctional', args, kwargs)
        self._tasks.append(ModifierTask('set_xcfunctional', args, kwargs))



"""
This module collects functions for calculating properties with the greens functions
calculated by Fleur. At the moment the following are implemented:

   * Calculating Heisenberg J_0 (spin stiffness) from onsite Green's functions
   * Calculating Heisenberg J_ij exchange constants from intersite Green's functions
   * Calculating the hybridization function from onsite Greens functions
"""
from __future__ import annotations

from masci_tools.util.typing import FileLike

from .greensfunction import GreensFunction, intersite_shells, intersite_shells_from_file
from masci_tools.io.common_functions import get_pauli_matrix

import numpy as np
import pandas as pd
from scipy import constants
from collections import defaultdict
from typing import Any

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type:ignore

#TODO:
# - Integration as one numpy operation
# - Parallelization/threading
# - multiple blocks/orbital contributions
# - custom decompositions (e.g. e2g/t2g)


def calculate_heisenberg_jij(
    hdffileORgreensfunctions: FileLike | list[GreensFunction],
    reference_atom: int,
    onsite_delta: np.ndarray,
    max_shells: int | None = None,
) -> pd.DataFrame:
    r"""
    Calculate the Heisenberg exchange constants form Green's functions using the formula

    .. math::
        J_{ij} = \frac{1}{4\pi} \mathrm{Im}\ \mathrm{Tr_L} \int_{-\infty}^{E_F}\!\mathrm{dz} \Delta_iG^\uparrow_{ij}(z)\Delta_jG^\downarrow_{ji}(z)

    :param hdffileORgreensfunctions: either pat/file-like object for the ``greensf.hdf`` file to use or list of :py:class:`~masci_tools.tools.greensfunction.GreensFunction`
    :param reference_atom: integer index of the atom to calculate the Jijs from
    :param onsite_delta: List of floats containing the onsite exchange splitting for each atom type and l-channel
    :param max_shells: optional int, if given only the first max_shells shells are constructed

    :returns: pandas DataFrame containing all the Jij constants
    """

    if isinstance(hdffileORgreensfunctions, list):
        shells = intersite_shells(hdffileORgreensfunctions, reference_atom, max_shells=max_shells)
    else:
        shells = intersite_shells_from_file(hdffileORgreensfunctions, reference_atom, max_shells=max_shells)

    jij_constants: dict[str, list[Any]] = defaultdict(list)

    for dist, g1, g2 in shells:

        dist = round(dist, 12)

        g1.to_global_frame()
        g2.to_global_frame()

        gij = g1.energy_dependence(both_contours=True, spin=1)
        gji = g2.energy_dependence(both_contours=True, spin=2)

        delta_square = onsite_delta[g1.atomType - 1, g1.l] * onsite_delta[g1.atomTypep - 1, g1.l]
        weights = np.array([g1.weights, -g1.weights.conj()]).T

        integral = np.einsum('zm,zijm,zjim->', weights, gij, gji)
        jij = 0.5 * 1 / (8.0 * np.pi * 1j) * delta_square * integral

        jij_constants['R'].append(dist)
        jij_constants['R_ij_x'].append(g1.atomDiff.tolist()[0])
        jij_constants['R_ij_y'].append(g1.atomDiff.tolist()[1])
        jij_constants['R_ij_z'].append(g1.atomDiff.tolist()[2])
        jij_constants['Atom i'].append(f"{g1.extras['atom_label']}({g1.extras['element']})")
        jij_constants['Atom j'].append(f"{g1.extras['atom_labelp']}({g1.extras['elementp']})")
        jij_constants['J_ij'].append(jij.real * 1000)  #Convert to meV

    return pd.DataFrame.from_dict(jij_constants)


def calculate_heisenberg_tensor(hdffileORgreensfunctions: FileLike | list[GreensFunction],
                                reference_atom: int,
                                onsite_delta: np.ndarray,
                                max_shells: int | None = None) -> pd.DataFrame:
    r"""
    Calculate the Heisenberg exchange tensor :math:`\mathbf{J}` from Green's functions using the formula

    .. math::
        J^{\alpha\beta}_{ij} = \frac{1}{4\pi} \mathrm{Im}\ \mathrm{Tr_L} \int_{-\infty}^{E_F}\!\mathrm{dz} \Delta_i\sigma_{\alpha}G_{ij}(z)\Delta_j\sigma_{\beta}G_{ji}(z)

    for all :math:`\alpha\ \mathrm{and}\ \beta=x,y,z`.

    :param hdffileORgreensfunctions: either pat/file-like object for the ``greensf.hdf`` file to use or list of :py:class:`~masci_tools.tools.greensfunction.GreensFunction`
    :param reference_atom: integer index of the atom to calculate the Jijs from
    :param onsite_delta: List of floats containing the onsite exchange splitting for each atom type and l-channel
    :param max_shells: optional int, if given only the first max_shells shells are constructed

    :returns: pandas DataFrame containing all the J_xx, J_xy, etc. constants
    """

    if isinstance(hdffileORgreensfunctions, list):
        shells = intersite_shells(hdffileORgreensfunctions, reference_atom, max_shells=max_shells)
    else:
        shells = intersite_shells_from_file(hdffileORgreensfunctions, reference_atom, max_shells=max_shells)

    jij_tensor: dict[str, list[Any]] = defaultdict(list)

    for dist, g1, g2 in shells:

        dist = round(dist, 12)

        g1.to_global_frame()
        g2.to_global_frame()
        gij = g1.energy_dependence(both_contours=True)
        gji = g2.energy_dependence(both_contours=True)

        delta_square = onsite_delta[g1.atomType - 1, g1.l] * onsite_delta[g1.atomTypep - 1, g1.l]
        weights = np.array([g1.weights, -g1.weights.conj()]).T

        jij_tensor['R'].append(dist)
        jij_tensor['R_ij_x'].append(g1.atomDiff.tolist()[0])
        jij_tensor['R_ij_y'].append(g1.atomDiff.tolist()[1])
        jij_tensor['R_ij_z'].append(g1.atomDiff.tolist()[2])
        jij_tensor['Atom i'].append(f"{g1.extras['atom_label']}({g1.extras['element']})")
        jij_tensor['Atom j'].append(f"{g1.extras['atom_labelp']}({g1.extras['elementp']})")

        for sigmai_str in ('x', 'y', 'z'):
            for sigmaj_str in ('x', 'y', 'z'):

                sigmai = get_pauli_matrix(sigmai_str)  #type: ignore[arg-type]
                sigmaj = get_pauli_matrix(sigmaj_str)  #type: ignore[arg-type]

                integral = np.einsum('zm,ab,zijbcm,cd,zjidam->', weights, sigmai, gij, sigmaj, gji)
                jij = 1 / 4 * 1 / (8.0 * np.pi * 1j) * delta_square * integral
                jij_tensor[f'J_{sigmai_str}{sigmaj_str}'].append(jij.real * 1000)  #Convert to meV

    return pd.DataFrame.from_dict(jij_tensor)


def decompose_jij_tensor(jij_tensor: pd.DataFrame, moment_direction: Literal['x', 'y', 'z']) -> pd.DataFrame:
    r"""
    Decompose the Heisenberg tensor as calculated by :py:func:`calculate_heisenberg_tensor()`
    into three parts

    - Isotropic :math:`J = \frac{1}{3} \mathrm{Tr}\left[\mathbf{J}\right]`
    - Symmetric traceless :math:`J_S = \frac{1}{2} \left(\mathbf{J} + \mathbf{J}^T\right) - J`
    - Antisymmetric :math:`J_A = \frac{1}{2} \left(\mathbf{J} - \mathbf{J}^T\right)`

    :param jij_tensor: Heisenberg tensor

    :returns: tuple of the three aforementioned components
    """

    if moment_direction == 'x':
        jij_tensor['J_ij'] = 1 / 2 * (jij_tensor['J_yy'] + jij_tensor['J_zz'])  #Isotropic
        jij_tensor['A_ij'] = 1 / 2 * (jij_tensor['J_yy'] - jij_tensor['J_zz'])  #Difference in diagonal
        jij_tensor['S_ij'] = 1 / 2 * (jij_tensor['J_yz'] + jij_tensor['J_zy'])  #Offdiagonal symmetric
        jij_tensor['D_ij'] = 1 / 2 * (jij_tensor['J_yz'] - jij_tensor['J_zy'])  #Offdiagonal asymmetric
    elif moment_direction == 'y':
        jij_tensor['J_ij'] = 1 / 2 * (jij_tensor['J_xx'] + jij_tensor['J_zz'])  #Isotropic
        jij_tensor['A_ij'] = 1 / 2 * (jij_tensor['J_xx'] - jij_tensor['J_zz'])  #Difference in diagonal
        jij_tensor['S_ij'] = 1 / 2 * (jij_tensor['J_xz'] + jij_tensor['J_zx'])  #Offdiagonal symmetric
        jij_tensor['D_ij'] = 1 / 2 * (jij_tensor['J_xz'] - jij_tensor['J_zx'])  #Offdiagonal asymmetric
    elif moment_direction == 'z':
        jij_tensor['J_ij'] = 1 / 2 * (jij_tensor['J_xx'] + jij_tensor['J_yy'])  #Isotropic
        jij_tensor['A_ij'] = 1 / 2 * (jij_tensor['J_xx'] - jij_tensor['J_yy'])  #Difference in diagonal
        jij_tensor['S_ij'] = 1 / 2 * (jij_tensor['J_xy'] + jij_tensor['J_yx'])  #Offdiagonal symmetric
        jij_tensor['D_ij'] = 1 / 2 * (jij_tensor['J_xy'] - jij_tensor['J_yx'])  #Offdiagonal asymmetric

    else:
        raise ValueError(f'Invalid direction: {moment_direction}')

    return jij_tensor


def heisenberg_reciprocal(qpoints: np.ndarray, jij_data: pd.DataFrame, entry: str = 'J_ij') -> np.ndarray:
    r"""
    Calculate the fourier transform of an entry for interaction constants

    Example for :math:`J_{ij}`

    .. math::
        J(\mathbf{q})\ =\ \sum_{ij} J_{ij} e^{i\mathbf{q}\cdot\mathbf{R}_{ij}}

    where :math:`\mathbf{R}_{ij}` is the connecting vector associated with the :math:`J_{ij}`

    :param qpoints: numpy array containing the coordinates of the qpoints
    :param jij_data: DataFrame generated by the above calculation functions
    :param entry: str of the entry to calculate

    :returns: numpy array containing the Fourier transform
    """
    interactions = jij_data[entry]
    r_vectors = np.array([jij_data['R_ij_x'], jij_data['R_ij_y'], jij_data['R_ij_z']]).T
    expo = np.exp(1j * np.einsum('qi,ki->qk', qpoints, r_vectors))

    interactions_reciprocal = np.einsum('k,qk->q', interactions, expo)
    return interactions_reciprocal


def calculate_heisenberg_j0(greensfunction: GreensFunction, onsite_delta: float, show: bool = False) -> float:
    r"""
    Calculate spin stiffness J_0 for the given green's function using the formula

    .. math::
        J_{0} = \frac{1}{4\pi} \mathrm{Im}\ \mathrm{Tr_L} \int_{-\infty}^{E_F}\!\mathrm{dz} \Delta\left(G^\uparrow(z)-G^\downarrow(z)\right) + \Delta^2G^\uparrow(z)G^\downarrow(z)

    :param greensfunction: :py:class:`~masci_tools.tools.greensfunction.GreensFunction` to use for the calculation
    :param onsite_delta: onsite exchange splitting to use for the calculation
    :param show: bool if True additional information about the used Greens functions is printed out

    :returns: the value of the spin stiffness in meV
    """

    g_up = greensfunction.energy_dependence(spin=1, both_contours=True)
    g_dn = greensfunction.energy_dependence(spin=2, both_contours=True)

    j0 = 0.0
    for weight, g_upz, g_dnz in zip(greensfunction.weights, g_up, g_dn):

        j0 += 1/(8.0*np.pi*1j) * (\
              weight *        (onsite_delta * (np.trace(g_upz[...,0])-np.trace(g_dnz[...,0])) \
                                + onsite_delta**2 * np.trace(g_upz[...,0].dot(g_dnz[...,0]))) \
             -weight.conj() * (onsite_delta * (np.trace(g_upz[...,1])-np.trace(g_dnz[...,1])) \
                                + onsite_delta**2 * np.trace(g_upz[...,1].dot(g_dnz[...,1]))))

    j0 = j0.real * 1000.0
    if show:
        print(f'Effective exchange Interaction: {j0:.4f} meV')
        print(f"Curie Temperature: {2.0/3.0*j0*1.0/constants.value('Boltzmann constant in eV/K')*1/1000.0:.4f} K"
              )  #1/1000 to convert to meV

    return j0


def calculate_hybridization(greensfunction: GreensFunction) -> np.ndarray:
    r"""
    Calculate the hybridization function as

    .. math::
        \Delta(z) = \frac{1}{2*l+1} \mathrm{Tr} G^{-1}(z)

    :returns: numpy array of the hybridization function
    """
    gz = greensfunction.trace_energy_dependence()
    delta = 1 / (2 * greensfunction.l + 1) * np.linalg.inv(gz)
    return delta.real



"""
Script to keep docstrings of XML setter functions and their corresponding methods
on the FleurXMLModifier in sync
Is used as a pre-commit hook
"""
import ast
import sys
from pathlib import Path

INDENT = 4
MASCI_TOOLS_PATH = Path(__file__).parent.parent / 'masci_tools'


def get_method_docstring(name, docstring, module):
    """
    Get the corresponding method docstring of a given XML setter function

    This changes three things.
        - Strip out references to the arguments handled by the FleurXMLModifier
        - Add two standardized lines explaining the actual effect of callind the method
          (append an entry to the tasks)
        - Indentation is adjusted to two levels (function -> method)
    """

    lines = [line for line in docstring.split('\n') \
                if all(x not in line for x in (':param xmltree:', ':param schema_dict:', ':param nmmplines:', ':returns'))]

    additional_lines = [
        f'Appends a :py:func:`~masci_tools.util.xml.{module}.{name}()` to',
        'the list of tasks that will be done on the xmltree.', ''
    ]
    if lines[0]:
        lines.insert(0, '')
    if lines[-1]:
        lines.append('')

    for line in reversed(additional_lines):
        lines.insert(1, line)

    while all(not line.strip() for line in lines[-2:]):
        lines.pop()
    lines = [2 * INDENT * ' ' + line if line.strip() else line.lstrip() for line in lines]
    lines[-1] = 2 * INDENT * ' '
    #Two levels of indentation have to be added since the docstrings go into methods
    return '\n'.join(lines)


def rewrite_docstrings(module_file, modifier_class_name, setters, modules):
    """
    Rewrite all the docstrings of the XMl setter methods of the
    FleurXMLModifier class to be in sync with their corresponding functions
    """

    with open(module_file, encoding='utf-8') as f:
        content = f.read()
        module = ast.parse(content)
    class_definitions = [node for node in module.body if isinstance(node, ast.ClassDef)]

    failed = False
    for class_def in class_definitions:
        if class_def.name != modifier_class_name:
            continue
        function_definitions = [node for node in class_def.body if isinstance(node, ast.FunctionDef)]
        for f in function_definitions:
            if f.name in setters:
                try:
                    docstring = get_method_docstring(f.name, setters[f.name], modules[f.name])
                except Exception as exc:  #pylint: disable=broad-except
                    print(f'Docstring generation failed for: {f.name} ({exc})')
                    failed = True
                    continue
                old_docstring = ast.get_docstring(f, clean=False)
                if old_docstring != docstring:
                    print(f'Rewriting docstring of method: {f.name}')
                    content = content.replace(old_docstring, docstring)

    with open(module_file, 'w', encoding='utf-8') as f:
        f.write(content)

    if failed:
        sys.exit(1)


def gather_setter_functions(module_names, collection_file):
    """
    Gather all setter functions that are imported in collect_xml_setters
    """

    setter_files = [MASCI_TOOLS_PATH / f'util/xml/{name}.py' for name in module_names]
    docstrings = {}

    for file in setter_files:
        with open(file, encoding='utf-8') as f:
            module = ast.parse(f.read())

        function_definitions = [node for node in module.body if isinstance(node, ast.FunctionDef)]
        for f in function_definitions:
            docstrings[f.name] = ast.get_docstring(f)

    with open(collection_file, encoding='utf-8') as f:
        module = ast.parse(f.read())

    imports = [node for node in module.body if isinstance(node, ast.ImportFrom)]

    collected_docstrings = {}
    modules = {}

    for import_stmt in imports:
        if import_stmt.module in module_names:
            for alias in import_stmt.names:
                collected_docstrings[alias.name] = docstrings[alias.name]
                modules[alias.name] = import_stmt.module

    return collected_docstrings, modules


if __name__ == '__main__':

    setter_module_names = ('xml_setters_names', 'xml_setters_nmmpmat', 'xml_setters_basic')
    setter_docstrings, setter_modules = gather_setter_functions(setter_module_names,
                                                                MASCI_TOOLS_PATH / 'util/xml/collect_xml_setters.py')
    rewrite_docstrings(MASCI_TOOLS_PATH / 'io/fleurxmlmodifier.py', 'FleurXMLModifier', setter_docstrings,
                       setter_modules)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
dispersionplot function for plotting KKR bandstructures (i.e. qdos) files
"""

import numpy as np


def load_data(p0='./',
              reload_data=False,
              nosave=False,
              atoms=None,
              ratios=False,
              atoms2=None,
              atoms3=None,
              ef=None,
              verbose=False):
    """load the qdos data"""
    from os import listdir
    from os.path import isdir
    from subprocess import check_output

    # deal with input of file handle instead of path (see plot_kkr of aiida_kkr)
    if not isinstance(p0, str):
        pathname_with_file = p0.name
        p0 = pathname_with_file.replace('/qdos.01.1.dat', '')  #dos.atom1
    # read in data
    if p0[-1] != '/':
        p0 += '/'
    # read EF if not given as input
    if ef is None:
        if 'potential' in listdir(p0):
            with open(p0 + 'potential', encoding='utf-8') as _f:
                ef = float(_f.readlines()[3].split()[1])
        else:
            ef = 0
    alat = float(
        check_output('grep ALATBASIS ' + p0 + 'inputcard', shell=True).decode('utf-8').split('=')[1].split()[0])
    a0 = 2 * np.pi / alat / 0.52918

    d2, d3 = None, None
    if reload_data or 'saved_data_dispersion.npy' not in np.sort(listdir(p0)):
        first = True
        first2 = True
        first3 = True
        if verbose:
            print('reading qdos')
        j = 0
        for i in np.sort(listdir(p0)):
            if 'qdos.' in i[:5] and not isdir(p0 + i):
                iatom = i.replace('qdos.', '').split('.')[0]
                if atoms is None or int(iatom) in atoms:
                    j += 1
                    if verbose:
                        print(j, p0, i)
                    tmp = np.loadtxt(p0 + i)
                    tmp[:, 2:5] = tmp[:, 2:5]
                    if first:
                        d = tmp
                        first = False
                    else:
                        d[:, 5:] += tmp[:, 5:]
                if ratios and (atoms2 is None or int(iatom) in atoms2):
                    j += 1
                    if verbose:
                        print(j, p0, i, 'atoms2=', atoms2)
                    tmp = np.loadtxt(p0 + i)
                    tmp[:, 2:5] = tmp[:, 2:5]
                    if first2:
                        d2 = tmp
                        first2 = False
                    else:
                        d2[:, 5:] += tmp[:, 5:]  # pylint: disable=unsupported-assignment-operation
                if (atoms3 is None or int(iatom) in atoms3) and ratios:
                    j += 1
                    if verbose:
                        print(j, p0, i, 'atoms3=', atoms3)
                    tmp = np.loadtxt(p0 + i)
                    tmp[:, 2:5] = tmp[:, 2:5]
                    if first3:
                        d3 = tmp
                        first3 = False
                    else:
                        d3[:, 5:] += tmp[:, 5:]  # pylint: disable=unsupported-assignment-operation
        if not nosave:
            if verbose:
                print('saving data')
            np.save(p0 + 'saved_data_dispersion', d)
    else:
        if verbose:
            print('loading data')
        d = np.load(p0 + 'saved_data_dispersion.npy')

    return a0, d, d2, d3, ef


def dispersionplot(  # pylint: disable=inconsistent-return-statements
        p0='./',
        data_all=None,  # alternative way, skips loading data with the load_data function
        totonly=True,
        s=20,
        ls_ef=':',
        lw_ef=1,
        units='eV_rel',
        noefline=False,
        color='',
        reload_data=False,
        clrbar=True,
        logscale=True,
        nosave=False,
        atoms=None,
        ratios=False,
        atoms2=None,
        noscale=False,
        newfig=False,
        cmap=None,
        alpha=1.0,
        qcomponent=-2,
        clims=None,
        xscale=1.,
        raster=True,
        atoms3=None,
        alpha_reverse=False,
        return_data=False,
        xshift=0.,
        yshift=0.,
        plotmode='pcolor',
        ptitle=None,
        ef=None,
        as_e_dimension=None,
        scale_alpha_data=False,
        shading='gouraud',
        verbose=False):
    """ plotting routine for qdos files - dispersion (E vs. q) """
    # import dependencies
    from numpy import log, abs, sum, sort, pi, shape, array  # pylint: disable=redefined-builtin
    from matplotlib.pyplot import figure, plot, axvline, scatter, axhline, xlabel, ylabel, title, colorbar, pcolormesh, cm, xlim, ylim, clim
    from os import listdir, getcwd
    from os.path import getctime
    from time import ctime
    from numpy import linspace
    from matplotlib.colors import ListedColormap

    if data_all is None:
        # load data from files
        a0, d, d2, d3, ef = load_data(p0=p0,
                                      reload_data=reload_data,
                                      nosave=nosave,
                                      atoms=atoms,
                                      ratios=ratios,
                                      atoms2=atoms2,
                                      atoms3=atoms3,
                                      ef=ef,
                                      verbose=verbose)
    else:
        # simply unpack the data_all input
        a0, d, d2, d3, ef = data_all

    if verbose:
        print('loaded/unpacked data')

    if cmap is None:
        cmap = cm.viridis

    if newfig:
        if verbose:
            print('open new figure')
        figure()

    if noscale:
        a0 = 1.

    d[:, 2:5] = d[:, 2:5] * a0
    if ratios:
        d2[:, 2:5] = d2[:, 2:5] * a0
        d3[:, 2:5] = d3[:, 2:5] * a0

    # set units and axis labels
    if 'eV' in units:
        d[:, 0] = d[:, 0] * 13.6
        d[:, 5:] = d[:, 5:] / 13.6
        ef = ef * 13.6
    if 'rel' in units:
        d[:, 0] = d[:, 0] - ef
        ef = 0
    if ratios:
        if 'eV' in units:
            d2[:, 5:] = d2[:, 5:] / 13.6
            d3[:, 5:] = d3[:, 5:] / 13.6

    ylab = r'E (Ry)'
    xlab = r'k'
    if units == 'eV':
        ylab = r'E (eV)'
    elif units == 'eV_rel':
        ylab = r'E-E_F (eV)'
    elif units == 'Ry_rel':
        ylab = r'E-E_F (Ry)'

    # plot dos
    if totonly:
        data = abs(sum(d[:, 5:], axis=1))
    else:
        data = abs(d[:, 5:])
    if logscale:
        data = log(data)
    x, y = xscale * sum(d[:, 2:5], axis=1), d[:, 0]
    if qcomponent == -2:
        el = len(set(d[:, 0]))  # pylint disable=unnecessary-comprehension
        if el == 1 and as_e_dimension is not None:
            y = d[:, 2 + as_e_dimension]
            el = len(set(y))
            ylab = r'k (1/Ang.)'
        x = array([list(range(len(d) // el)) for j in range(el)])
    elif qcomponent != -1:
        x = xscale * d[:, 2:5][:, qcomponent]

    if xshift != 0:
        x += xshift

    if ratios:
        data = abs(sum(d[:, 5:], axis=1))
        data2 = abs(sum(d2[:, 5:], axis=1))
        data = (data - data2) / (data + data2)

    if yshift != 0:
        y += yshift

    if scale_alpha_data:
        dtmp = data.copy()
        dtmp = linspace(dtmp.min(), dtmp.max(), 1000)
        colors = cmap(dtmp)
        dtmp = dtmp - dtmp.min()
        dtmp = dtmp / dtmp.max()
        colors[:, -1] = alpha * (dtmp)
        cmap = ListedColormap(colors)
        alpha = 1.
    if ratios and atoms3 is not None:
        colors = cmap(data / data.max())
        colors[:, -1] = abs(sum(d3[:, 5:], axis=1)) / abs(sum(d3[:, 5:], axis=1)).max()
        if alpha_reverse:
            colors[:, -1] = 1 - colors[:, -1]
    if ratios:
        if plotmode == 'scatter':
            scatter(x, y, s=s, lw=0, c=colors, cmap=cmap, rasterized=raster)
        else:
            lx = len(set(x.reshape(-1)))
            ly = len(set(y.reshape(-1)))
            pcolormesh(
                x.reshape(ly, lx),
                y.reshape(ly, lx),
                data.reshape(ly, lx),
                cmap=cmap,
                rasterized=raster,
                edgecolor='face',
                linewidths=0.0001,
                shading=shading,
            )
    else:
        if plotmode == 'scatter':
            scatter(x, y, c=data, s=s, lw=0, cmap=cmap, alpha=alpha, rasterized=raster)
        else:
            lx = len(set(x.reshape(-1)))
            ly = len(set(y.reshape(-1)))
            pcolormesh(
                x.reshape(ly, lx),
                y.reshape(ly, lx),
                data.reshape(ly, lx),
                cmap=cmap,
                rasterized=raster,
                edgecolor='face',
                linewidths=0.0001,
                shading=shading,
            )
    if clims is not None:
        clim(clims[0], clims[1])
    if clrbar:
        colorbar()

    # plot fermi level
    if not noefline:
        if color == '':
            axhline(ef, ls=ls_ef, lw=lw_ef, color='grey')
        else:
            axhline(ef, color=color, ls=ls_ef, lw=lw_ef)

    # set axis labels
    xlabel(xlab)
    ylabel(ylab)

    # set x and y limits
    xlim(x.min(), x.max())
    ylim(y.min(), y.max())

    # print path to title
    if totonly and ptitle is None:
        title(getcwd())
    else:
        title(ptitle)

    if return_data:
        data = abs(sum(d[:, 5:], axis=1))

        if ratios:
            data2 = abs(sum(d2[:, 5:], axis=1))
            data3 = abs(sum(d3[:, 5:], axis=1))
            return x, y, data, data2, data3

        return x, y, data



"""
Collection of routines to be reused in plotting routines
"""
from __future__ import annotations

import numpy as np

from .data import PlotData


def exclude_points(plot_data: PlotData,
                   *data_keys: str,
                   limits: dict[str, tuple[float, float]] | None,
                   padding: float = 0.1) -> None:
    """
    Exclude points outside the given limits

    :param plot_data: PlotData instance containing all the data for plots
    :param data_keys: str of the keys to consider for excluding points
    :param limits: dict of the set plot limits
    :param padding: float, determines how far beyond the plot limits a point has to lie to be excluded (default 10%)
    """

    if limits is None:
        return

    combined_mask = None
    for data_key in data_keys:

        if data_key in limits:
            data_limits = limits[data_key]

            #Add padding to both sides of the limits
            data_limits = data_limits[0] - padding * (1 + abs(data_limits[0])), data_limits[1] + padding * (
                1 + abs(data_limits[1]))

            mask = plot_data.get_mask(
                lambda x, data_limits=tuple(data_limits): np.logical_and(x > data_limits[0], x < data_limits[1]),
                data_key=data_key)

            if combined_mask is None:
                combined_mask = mask
            else:
                combined_mask = [x & y for x, y in zip(mask, combined_mask)]

    if combined_mask is not None:
        plot_data.mask_data(combined_mask)


def mpl_single_line_or_area(axis,
                            entry,
                            source,
                            area=False,
                            area_vertical=False,
                            area_enclosing_line=True,
                            advance_color_cycle=False,
                            area_line_alpha=1.0,
                            **kwargs):
    """
    Create a scatterplot, lineplot or area plot for the given entry
    on the matplotlib axis

    :param axis: Axes to plot on
    :param entry: namedtuple of the entries to plot
    :param source: mapping containing the data to plot
    :param area: bool, if True fill_betweenx/y will be used to create an area plot
    :param area_vertical: bool, if True fill_betweeny will be used
    :param area_enclosing_line: bool if True the area plot will have another
                                line plot around the edge
    :param advance_color_cycle: bool, if True the matplotlib color cycle will be advanced
                                no matter what else is specified
    :param area_line_alpha: if an area plot is done this is the alpha parameter (transparency)

    Kwargs are passed to the respective plotting routines
    """
    if any(key not in entry._fields for key in ('x', 'y')):
        raise ValueError('Entry has to contain x and y fields')

    if area and 'shift' not in entry._fields:
        raise ValueError('For area plots entry has to contain shift fields')

    yerr = entry.yerr if 'yerr' in entry._fields else None
    xerr = entry.xerr if 'xerr' in entry._fields else None

    if area:
        linecolor = kwargs.pop('area_linecolor', None)
        #Workaround for https://github.com/JuDFTteam/masci-tools/issues/129
        #fill_between does not advance the color cycle messing up the following colors
        if kwargs.get('color') is None:
            #This is not ideal but it is the only way I found
            #of accessing the state of the color cycle
            kwargs['color'] = axis._get_lines.get_next_color()  #pylint: disable=protected-access

        if area_vertical:
            result = axis.fill_betweenx(entry.y, entry.x, x2=entry.shift, data=source, **kwargs)
        else:
            result = axis.fill_between(entry.x, entry.y, y2=entry.shift, data=source, **kwargs)

        kwargs['alpha'] = area_line_alpha
        kwargs['label'] = None
        kwargs['color'] = linecolor or result.get_facecolor()[0]

        if area_enclosing_line:
            axis.errorbar(entry.x, entry.y, yerr=yerr, xerr=xerr, data=source, **kwargs)
    else:
        if kwargs.get('color') is not None and advance_color_cycle:
            axis._get_lines.get_next_color()  #pylint: disable=protected-access
        axis.errorbar(entry.x, entry.y, yerr=yerr, xerr=xerr, data=source, **kwargs)

    return axis


def get_special_kpoint_ticks(kpoints: list[tuple[str, float]], math_mode: str = '$') -> tuple[list[float], list[str]]:
    r"""
    Process the high symmetry kpoints and ggf. replace with appropiate latex symbol

    - Gamma/G is replaced with $\Gamma$

    :param kpoints: list of tuples with label and position of the points
    :param math_mode: Determines the symbol to enter math mode in latex

    :returns: Ticks and their respective labels
    """
    if kpoints is None:
        kpoints = []

    ticks = []
    ticklabels = []
    for label, position in kpoints:
        if label.lower() in ('gamma', 'g'):
            label = rf'{math_mode}\Gamma{math_mode}'
        ticklabels.append(label)
        ticks.append(position)

    return ticks, ticklabels



"""
This module contains a helper class for creating Elements compatible with a given
input version of fleur from scratch

Example of creating a kpointlist

.. code-block:: python
    from masci_tools.util.xml.builder import FleurElementMaker

    E = FleurElementMaker.fromVersion('0.35')

    new_kpointset = E.kpointlist(
        *(
            E.kpoint(kpoint, weight=weight, label=special_labels[indx]) if indx in special_labels else
            E.kpoint(kpoint, weight=weight) for indx, (kpoint, weight) in enumerate(zip(kpoints, weights))
        ),
        name=name,
        count=nkpts,
        type=kpoint_type)

"""
from __future__ import annotations

from lxml.builder import ElementMaker  #pylint: disable=no-name-in-module
from lxml import etree

from masci_tools.util.xml.converters import convert_to_xml
from masci_tools.io.parsers.fleur_schema import InputSchemaDict, SchemaDict

from typing import Any, Callable, Iterable


class FleurElementMaker(ElementMaker):  #type: ignore[misc]
    """
    Class to create XML elements for fleur XML files
    utilizing the schema dict.
    For more information refer to the lxml documentation
    on :py:class:`lxml.builder.ElementMaker`

    :param schema_dict: InputSchemaDict to use for validation/conversion
    :param typemap: dict to add functions to handle given types
    :param namespace: str of the xml namespace to use
    :param nsmap: mapping between namespace names and uris
    :param makeelement: optional function to construct XML elements

    .. code-block:: python
        from masci_tools.util.xml.builder import FleurElementMaker

        E = FleurElementMaker.fromVersion('0.35')

        new_kpointset = E.kpointlist(
            *(
                E.kpoint(kpoint, weight=weight, label=special_labels[indx]) if indx in special_labels else
                E.kpoint(kpoint, weight=weight) for indx, (kpoint, weight) in enumerate(zip(kpoints, weights))
            ),
            name=name,
            count=nkpts,
            type=kpoint_type)

    """

    @classmethod
    def fromVersion(cls, version: str, **kwargs: Any) -> FleurElementMaker:
        """
        Create a FleurElementMaker for the given input version

        :param version: version string
        """

        schema_dict = InputSchemaDict.fromVersion(version)
        return cls(schema_dict, **kwargs)

    def __init__(self,
                 schema_dict: SchemaDict,
                 typemap: dict | None = None,
                 namespace: str | None = None,
                 nsmap: dict[str, str] | None = None,
                 makeelement: Callable | None = None) -> None:
        super().__init__(typemap=typemap, namespace=namespace, nsmap=nsmap, makeelement=makeelement)
        self.schema_dict = schema_dict

    def __call__(self,
                 tag: str,
                 *children: Any,
                 contains: str | Iterable[str] | None = None,
                 not_contains: str | Iterable[str] | None = None,
                 **attrib: Any) -> etree._Element:
        """
        Construct a element with the given tag, text children tags and attributes

        Args are used to set child elements if they are `etree._Element` or text of the given tag
        Kwargs are used to set attributes on the given tag

        :param tag: str of the tag name. Is passed to :py:meth:`masci_tools.io.parsers.fleur_schema.InputSchemaDict.tag_xpath()`
        :param contains: specification of the tag name. Is passed to :py:meth:`masci_tools.io.parsers.fleur_schema.InputSchemaDict.tag_xpath()`
        :param not_contains: specification of the tag name. Is passed to :py:meth:`masci_tools.io.parsers.fleur_schema.InputSchemaDict.tag_xpath()`

        :returns: `etree._Element` constructed from the given information
        """
        tag_info = self.schema_dict.tag_info(tag, contains=contains, not_contains=not_contains)
        tag = tag_info['name']  #original case

        children_converted = list(children)
        for indx, child in enumerate(children_converted):
            if isinstance(child, str) or etree.iselement(child):
                continue
            value, success = convert_to_xml(child, self.schema_dict, tag, text=True)
            if not success:
                raise ValueError(
                    f'Failed to convert value `{value}` for text of {tag}. The types are probably not right')
            children_converted[indx] = value

        #convert attributes
        attrib_converted = {}
        for key, value in attrib.items():

            if key not in tag_info['attribs']:
                raise KeyError(f'The attribute {key} is not allowed for the given tag {tag}')

            attrib_name = tag_info['attribs'].original_case[key]
            if not isinstance(value, str):
                value, success = convert_to_xml(value, self.schema_dict, attrib_name)
                if not success:
                    raise ValueError(
                        f'Failed to convert value `{value}` for attribute {attrib_name}. The types are probably not right'
                    )

            attrib_converted[attrib_name] = value

        return super().__call__(tag, *children_converted, **attrib_converted)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains functions to load an fleur inp.xml file, parse it with a schema
and convert its content to a dict
"""
from __future__ import annotations

from lxml import etree

from masci_tools.io.fleur_xml import get_constants, load_inpxml
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.xml.converters import convert_from_xml
from masci_tools.util.schema_dict_util import evaluate_attribute
from masci_tools.util.logging_util import DictHandler
from masci_tools.util.typing import XMLFileLike
import logging
from typing import Any
from masci_tools.io.parsers.fleur_schema import InputSchemaDict, EMPTY_TAG_INFO


def inpxml_parser(inpxmlfile: XMLFileLike,
                  parser_info_out: dict[str, Any] | None = None,
                  strict: bool = False,
                  debug: bool = False,
                  base_url: str | None = None) -> dict[str, Any]:
    """
    Parses the given inp.xml file to a python dictionary utilizing the schema
    defined by the version number to validate and correctly convert to the dictionary

    :param inpxmlfile: either path to the inp.xml file, opened file handle (in bytes modes i.e. rb)
                       or a xml etree to be parsed
    :param parser_info_out: dict, with warnings, info, errors, ...
    :param strict: bool if True  and no parser_info_out is provided any encountered error will immediately be raised

    :return: python dictionary with the parsed inp.xml

    :raises ValueError: If the validation against the schema failed, or an irrecoverable error
                        occurred during parsing
    :raises FileNotFoundError: If no Schema file for the given version was found

    """

    __parser_version__ = '0.3.0'
    logger: logging.Logger | None = logging.getLogger(__name__)

    if strict:
        logger = None

    parser_log_handler = None
    if logger is not None:
        if parser_info_out is None:
            parser_info_out = {}

        logging_level = logging.INFO
        if debug:
            logging_level = logging.DEBUG
        logger.setLevel(logging_level)

        parser_log_handler = DictHandler(parser_info_out,
                                         WARNING='parser_warnings',
                                         ERROR='parser_errors',
                                         INFO='parser_info',
                                         DEBUG='parser_debug',
                                         CRITICAL='parser_critical',
                                         ignore_unknown_levels=True,
                                         level=logging_level)

        logger.addHandler(parser_log_handler)

    if logger is not None:
        logger.info('Masci-Tools Fleur inp.xml Parser v%s', __parser_version__)

    xmltree, schema_dict = load_inpxml(inpxmlfile, logger=logger, base_url=base_url)
    actual_inp_version = evaluate_attribute(xmltree, schema_dict, 'fleurInputVersion', logger=logger)
    ignore_validation = schema_dict['inp_version'] != actual_inp_version

    xmltree, _ = clear_xml(xmltree)
    root = xmltree.getroot()

    constants = get_constants(root, schema_dict, logger=logger)

    try:
        schema_dict.validate(xmltree, logger=logger)
    except ValueError as err:
        if not ignore_validation:
            if logger is not None:
                logger.exception(err)
            raise

    inp_dict = inpxml_todict(root, schema_dict, constants, logger=logger)

    if parser_log_handler is not None:
        if logger is not None:
            logger.removeHandler(parser_log_handler)

    return inp_dict


def inpxml_todict(parent: etree._Element,
                  schema_dict: InputSchemaDict,
                  constants: dict[str, float],
                  omitted_tags: bool = False,
                  base_xpath: str | None = None,
                  logger: logging.Logger | None = None) -> dict[str, Any]:
    """
    Recursive operation which transforms an xml etree to
    python nested dictionaries and lists.
    Decision to add a list is if the tag name is in the given list tag_several

    :param parent: some xmltree, or xml element
    :param schema_dict: structure/layout of the xml file in python dictionary
    :param constants: dict with all the defined constants
    :param omitted_tags: switch. If True only a list of the contained tags is returned
                         Used to omit useless tags like e.g ['atomSpecies']['species'][3]
                         becomes ['atomSpecies'][3]
    :param base_xpath: str, keeps track of the place in the inp.xml currently being processed
    :param parser_info_out: dict, with warnings, info, errors, ...

    :return: a python dictionary
    """
    #These keys have to never appear as an attribute/tag name
    #The underscores should guarantee that
    _TEXT_PLACEHOLDER = '__text__'
    _OMIT_PLACEHOLDER = '__omit__'

    #Check if this is the first call to this routine
    if base_xpath is None:
        base_xpath = f'/{parent.tag}'

    content: dict[str, Any] = {}
    # Now we have to convert lazy fortran style into pretty things for the Database
    for key, value in parent.items():
        attrib_name, value = str(key), str(value)
        if attrib_name in schema_dict['attrib_types']:
            content[attrib_name], suc = convert_from_xml(value,
                                                         schema_dict,
                                                         attrib_name,
                                                         text=False,
                                                         constants=constants,
                                                         logger=logger)
            if not suc and logger is not None:
                logger.warning("Failed to convert attribute '%s' Got: '%s'", attrib_name, value)

    # has text, but we don't want all the '\n' s and empty strings in the database
    if parent.text and parent.text.strip() != '':

        if parent.tag not in schema_dict['text_tags']:
            if logger is not None:
                logger.error('Something is wrong in the schema_dict: %s is not in text_tags, but it has text',
                             parent.tag)
            raise ValueError(
                f'Something is wrong in the schema_dict: {parent.tag} is not in text_tags, but it has text')

        converted_text, suc = convert_from_xml(str(parent.text),
                                               schema_dict,
                                               parent.tag,
                                               text=True,
                                               constants=constants,
                                               logger=logger)

        if not suc and logger is not None:
            logger.warning("Failed to text of '%s' Got: '%s'", parent.tag, parent.text)

        content[_TEXT_PLACEHOLDER] = converted_text

    tag_info = schema_dict['tag_info'].get(base_xpath, EMPTY_TAG_INFO)
    for element in parent:

        child_content = inpxml_todict(element,
                                      schema_dict,
                                      constants,
                                      base_xpath=f'{base_xpath}/{element.tag}',
                                      omitted_tags=element.tag in schema_dict['omitt_contained_tags'],
                                      logger=logger)

        if _OMIT_PLACEHOLDER in child_content:
            #We knoe that there is only one key here
            child_content = child_content.pop(_OMIT_PLACEHOLDER)

        tag_name = element.tag
        if omitted_tags:
            tag_name = _OMIT_PLACEHOLDER

        if element.tag in tag_info['several']\
           and _TEXT_PLACEHOLDER in child_content:
            #The text is stored under the name of the tag
            text_value = child_content.pop(_TEXT_PLACEHOLDER)
            content.setdefault(tag_name, []).append(text_value)
            child_tag_info = schema_dict['tag_info'].get(f'{base_xpath}/{element.tag}', EMPTY_TAG_INFO)
            for key, value in child_content.items():
                if key not in child_tag_info['optional_attribs']:
                    #All required attributes are stored as lists
                    if key in content and \
                       not isinstance(content[key], list):  #Key seems to be defined already
                        if logger is not None:
                            logger.error('%s cannot be extracted to the next level', key)
                        raise ValueError(f'{key} cannot be extracted to the next level')
                    content.setdefault(key, []).append(value)
                else:
                    #All optional attributes are stored as dicts pointing to the text
                    content.setdefault(key, {})[value] = text_value
        elif element.tag in tag_info['several']:
            content.setdefault(tag_name, []).append(child_content)
        elif _TEXT_PLACEHOLDER in child_content:
            content[tag_name] = child_content.pop(_TEXT_PLACEHOLDER)
        else:
            content[tag_name] = child_content

    return content
